
import logging
from os import getenv

import aiohttp

log = logging.getLogger(__name__)


class APIException(Exception):
    pass


class NotCreatedException(APIException):
    pass


class BadStatusException(APIException):
    pass


def normalize_api_host(host: str):

    if not host:
        raise ValueError("host cannot be empty")

    if not host.endswith("/"):
        host += "/"

    return host


class API:
    """Interactions with the API."""

    API_HOST = normalize_api_host(getenv("API_HOST"))
    API_ENDPOINT = API_HOST + "api/"
    
    API_EXTERNAL_HOST = normalize_api_host(getenv("API_EXTERNAL_HOST"))
    API_EXTERNAL_ENDPOINT = API_EXTERNAL_HOST + "api/"

    RSS_FEED_ENDPOINT = API_ENDPOINT + "rssfeed/"
    FEED_CHANNEL_ENDPOINT = API_ENDPOINT + "feedchannel/"

    SUBSCRIPTION_ENDPOINT = API_ENDPOINT + "subscription/"
    CHANNEL_ENDPOINT = SUBSCRIPTION_ENDPOINT + "channel/"
    TRACKED_ENDPOINT = API_ENDPOINT + "tracked/"

    def __init__(self, api_token: str, session: aiohttp.ClientSession):
        log.debug("API session initialised")
        self.session = session
        self.token_headers = {"Authorization": f"Token {api_token}"}

    async def make_request(self, method: str, url: str, **kwargs) -> dict:
        """Make a request to the given API endpoint.

        Args:
            method (str): The request method to use, examples: GET, POST, DELETE...
            url (str): The API endpoint to request to.
            **kwargs: Passed into self.session.request. 

        Returns:
            dict: Dictionary containing status code, json or text.
        """

        log.info("making request to '%s' '%s'", url, method)

        async with self.session.request(method, url, headers=self.token_headers, **kwargs) as response:
            response.raise_for_status()
            try:
                json = await response.json()
                text = None
            except aiohttp.ContentTypeError:
                json = None
                text = await response.text()

            status = response.status

        return {"json": json, "text": text, "status": status}

    async def _post_data(self, url: str, data: dict | aiohttp.FormData) -> dict:
        return await self.make_request(
            method="POST",
            url=url,
            data=data
        )

    async def _put_data(self, url: str, data: dict | aiohttp.FormData) -> dict:
        return await self.make_request(
            method="PUT",
            url=url,
            data=data
        )

    async def _get_one(self, url: str) -> dict:
        data = await self.make_request(
            method="GET",
            url=url
        )
        return data["json"]

    async def _get_many(self, url: str, filters: dict) -> tuple[list[dict], int]:
        data = await self.make_request(
            method="GET",
            url=url,
            params=filters
        )
        content = data["json"]
        return content["results"], content["count"]

    async def _delete(self, url: str) -> None:
        await self.make_request(
            method="DELETE",
            url=url
        )

    async def get_subscriptions(self, **filters) -> tuple[list[dict], int]:
        """
        Get multiple subscriptions.
        """

        log.debug("getting multiple subscriptions")

        return await self._get_many(self.API_ENDPOINT + "subscription/", filters)

    async def get_subscription_channels(self, **filters) -> tuple[list[dict], int]:
        """
        Get many subscription channels.
        """

        log.debug("getting multiple channels")

        return await self._get_many(self.API_ENDPOINT + "subchannel/", filters)

    async def get_guild_settings(self, **filters) -> tuple[list[dict], int]:
        """
        Get many guild settings.
        """

        log.debug("getting multiple guild settings")

        return await self._get_many(self.API_ENDPOINT + "guild-settings/", filters)

    async def create_tracked_content(self, **data) -> dict:
        """
        Create an instance of tracked content.
        """

        log.debug("creating tracked content")

        return await self._post_data(self.API_ENDPOINT + "tracked-content/", data)

    async def get_tracked_content(self, **filters) -> tuple[list[dict], int]:
        """
        Return an instance of tracked content.
        """

        log.debug("getting tracked content")

        return await self._get_many(self.API_ENDPOINT + f"tracked-content/", filters)

    async def get_filter(self, filter_id: int) -> dict:
        """
        Get an instance of Filter.
        """

        log.debug("getting a filter")

        return await self._get_one(f"{self.API_ENDPOINT}filter/{filter_id}")
