import re
import logging

from rapidfuzz import fuzz

log = logging.getLogger(__name__)

def _split_match(_match):
    """
    Splits the match to individual keywords, getting rid of unnecessary
    spaces and grouping quoted words together.

    Example:
    '  some random  words "with   quotes  " and   spaces'
        ==>
    ["some", "random", "words", "with+quotes", "and", "spaces"]
    """
    findterms = re.compile(r'"([^"]+)"|(\S+)').findall
    normspace = re.compile(r"\s+").sub
    return [
        # normspace(" ", (t[0] or t[1]).strip()).replace(" ", r"\s+")
        re.escape(normspace(" ", (t[0] or t[1]).strip())).replace(r"\ ", r"\s+")
        for t in findterms(_match)
    ]

def _match_any(_match: str, matching_to: str, **search_kwargs) -> bool:
    for word in _split_match(_match):
        if re.search(rf"\b{word}\b", matching_to, **search_kwargs):
            return True

    return False

def _match_all(_match: str, matching_to: str, **search_kwargs) -> bool:
    for word in _split_match(_match):
        if not re.search(rf"\b{word}\b", matching_to, **search_kwargs):
            return False

    return True

def _match_literal(_match: str, matching_to: str, **search_kwargs) -> bool:
    return bool(
        re.search(
            rf"\b{re.escape(_match)}\b",
            matching_to,
            **search_kwargs,
        ),
    )

def _match_regex(_match: str, matching_to: str, **search_kwargs) -> bool:
    try:
        return bool(re.search(
            re.compile(_match, **search_kwargs),
            matching_to,
        ))
    except re.error as err:
        log.error(err)
        return False

def _match_fuzzy(_match: str, matching_to: str, is_insensitive: bool) -> bool:
    _match = re.sub(r"[^\w\s]", "", _match)
    text = re.sub(r"[^\w\s]", "", matching_to)
    if is_insensitive:
        _match = _match.lower()
        text = text.lower()

    return fuzz.partial_ratio(_match, text, score_cutoff=90)

def match_text(_filter: dict, matching_to: str) -> bool:
    search_kwargs = {}

    algorithm = _filter["matching_algorithm"]
    _match = _filter["match"]
    is_whitelist = _filter["is_whitelist"]
    is_insensitive = _filter["is_insensitive"]

    # If the matching pattern is empty
    if not _match.strip():
        return False

    if is_insensitive:
        search_kwargs = {"flags": re.IGNORECASE}

    log.debug(
        "matching algorithm '%s', whitelist? '%s', insensitive? '%s'",
        algorithm, is_whitelist, is_insensitive
    )

    match algorithm:
        case 0:
            return False

        case 1:  # Any
            return _match_any(_match, matching_to, **search_kwargs)

        case 2:  # All
            return _match_all(_match, matching_to, **search_kwargs)

        case 3:  # Exact Match
            return _match_literal(_match, matching_to, **search_kwargs)

        case 4:  # Regular Expression
            return _match_regex(_match, matching_to, **search_kwargs)

        case 5:  # Fuzzy Match
            return _match_fuzzy(_match, matching_to, is_insensitive)

        case _:
            return False
